# -*- coding: utf-8 -*-
"""Collection of executor classes for various submodes."""

# Standard Libraries
import re

# Third Party Libraries
from spyder.config.manager import CONF

# Project Libraries
from spyder_okvim.executor.executor_base import (
    FUNC_INFO,
    RETURN_EXECUTOR_METHOD_INFO,
    ExecutorSubBase,
)
from spyder_okvim.executor.executor_easymotion import ExecutorEasymotion
from spyder_okvim.executor.executor_surround import (
    ExecutorAddSurround,
    ExecutorChangeSurround,
    ExecutorDeleteSurround,
)
from spyder_okvim.spyder.config import CONF_SECTION


class ExecutorSubMotion_i(ExecutorSubBase):
    """Submode of iw, i( ..."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

        self.has_zero_cmd = False

    def __call__(self, txt):
        """Parse command and execute."""
        self.update_input_cmd_info(None, None, txt)
        self.vim_status.sub_mode = None

        if txt in "wW{}[]()'\"bBc":
            if txt == "w":
                motion_info = self.helper_motion.i_w(self.parent_num[-1])
            elif txt == "W":
                motion_info = self.helper_motion.i_W(self.parent_num[-1])
            elif txt in "'\"":
                motion_info = self.helper_motion.i_quote(txt)
            elif txt == "c":
                motion_info = self.helper_motion.i_cell(self.parent_num[-1])
            else:
                txt = txt.replace("b", "(")
                txt = txt.replace("B", "{")
                motion_info = self.helper_motion.i_bracket(self.parent_num[-1], txt)

            if motion_info.sel_start is None or motion_info.sel_end is None:
                return True

            return self.process_return(self.execute_func_deferred(motion_info))

        return True


class ExecutorSubMotion_a(ExecutorSubBase):
    """Submode of aw, a( ..."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

        self.has_zero_cmd = False

    def __call__(self, txt):
        """Parse command and execute."""
        self.update_input_cmd_info(None, None, txt)
        self.vim_status.sub_mode = None

        if txt in "w{}[]()'\"bB":
            if txt in "w":
                motion_info = self.helper_motion.a_w(self.parent_num[-1])
            elif txt in "'\"":
                motion_info = self.helper_motion.a_quote(txt)
            else:
                txt = txt.replace("b", "(")
                txt = txt.replace("B", "{")
                motion_info = self.helper_motion.a_bracket(self.parent_num[-1], txt)

            if motion_info.sel_start is None or motion_info.sel_end is None:
                return True

            return self.process_return(self.execute_func_deferred(motion_info))

        return True


class ExecutorSubMotion(ExecutorSubBase):
    """Submode of motion."""

    def __init__(self, vim_status):
        super().__init__(vim_status)

        self.has_zero_cmd = True

        self.cmds = "/nNailhkjHML$^wWbBegG%fFtT;,`' \b\r*#z[]"
        cmds = "".join(re.escape(c) for c in self.cmds)
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))
        self.executor_sub_sub_g = ExecutorSubSubCmd_g(vim_status)
        self.executor_sub_f_t = ExecutorSubCmd_f_t(vim_status)
        self.executor_sub_motion_i = ExecutorSubMotion_i(vim_status)
        self.executor_sub_motion_a = ExecutorSubMotion_a(vim_status)
        self.executor_sub_alnum = ExecutorSubCmd_alnum(vim_status)
        self.executor_sub_search = ExecutorSearch(vim_status)
        self.executor_sub_easymotion = ExecutorEasymotion(vim_status)
        self.executor_sub_sneak = ExecutorSubCmdSneak(vim_status)
        self.executor_sub_opensquarebracekt = ExecutorSubCmd_opensquarebracket(
            vim_status
        )
        self.executor_sub_closesquarebracekt = ExecutorSubCmd_closesquarebracket(
            vim_status
        )

    def __call__(self, txt):
        """Parse command and execute."""
        if txt[-1] == self.vim_status.input_cmd.cmd[-1]:
            # Handle the case if the input is the same with previous input.
            cmd = txt[-1]
            if len(txt) == 1:
                num_str = ""
                num = 1
            else:
                num_str = txt[:-1]
                num = int(num_str)
            self.update_input_cmd_info(num_str, cmd, txt)

            self.vim_status.sub_mode = None

            return self.process_return(self.same_ch_input(num, num_str))
        else:
            return super().__call__(txt)

    def same_ch_input(self, num=1, num_str=""):
        """Handle the case if the input is the same with previous input."""
        num = num * self.parent_num[0] - 1
        motion_info = self.helper_motion.j(num)
        return self.execute_func_deferred(motion_info)

    def zero(self, num=1, num_str=""):
        """Go to the 1st character of the line and execute."""
        motion_info = self.helper_motion.zero()
        return self.execute_func_deferred(motion_info)

    def l(self, num=1, num_str=""):
        """Go to [num] characters to the right and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.l(num)
        return self.execute_func_deferred(motion_info)

    def h(self, num=1, num_str=""):
        """Go to [num] characters to the left and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.h(num)
        return self.execute_func_deferred(motion_info)

    def k(self, num=1, num_str=""):
        """Go to [num] lines upward linewise and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.k(num)
        return self.execute_func_deferred(motion_info)

    def j(self, num=1, num_str=""):
        """Go to [num] lines downward linewise and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.j(num)
        return self.execute_func_deferred(motion_info)

    def H(self, num=1, num_str=""):
        """Go to top of window(linewise) and execute."""
        motion_info = self.helper_motion.H()
        return self.execute_func_deferred(motion_info)

    def M(self, num=1, num_str=""):
        """Go to middle line of window(linewise) and execute."""
        motion_info = self.helper_motion.M()
        return self.execute_func_deferred(motion_info)

    def L(self, num=1, num_str=""):
        """Go to bottom line of window(linewise) and execute."""
        motion_info = self.helper_motion.L()
        return self.execute_func_deferred(motion_info)

    def dollar(self, num=1, num_str=""):
        """Go to the end of the line and execute."""
        motion_info = self.helper_motion.dollar()
        return self.execute_func_deferred(motion_info)

    def caret(self, num=1, num_str=""):
        """Go to the non-blank character of the line and execute."""
        motion_info = self.helper_motion.caret()
        return self.execute_func_deferred(motion_info)

    def w(self, num=1, num_str=""):
        """Move forward [num] words and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.w(num)
        return self.execute_func_deferred(motion_info)

    def W(self, num=1, num_str=""):
        """Move forward [num] WORDs and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.W(num)
        return self.execute_func_deferred(motion_info)

    def b(self, num=1, num_str=""):
        """Move backward [num] words and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.b(num)
        return self.execute_func_deferred(motion_info)

    def B(self, num=1, num_str=""):
        """Move backward [num] WORDS and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.B(num)
        return self.execute_func_deferred(motion_info)

    def e(self, num=1, num_str=""):
        """Move to the end of [num] words and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.e(num)
        return self.execute_func_deferred(motion_info)

    def G(self, num=1, num_str=""):
        """Go to [num] line and execute."""
        num = num * self.parent_num[0]
        num_str = num_str or self.parent_num_str[0]
        motion_info = self.helper_motion.G(num, num_str)
        return self.execute_func_deferred(motion_info)

    def g(self, num=1, num_str=""):
        """Execute submode of g."""
        executor_sub = self.executor_sub_sub_g
        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def opensquarebracket(self, num=1, num_str=""):
        """Enter [ submode."""
        executor_sub = self.executor_sub_opensquarebracekt
        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def closesquarebracket(self, num=1, num_str=""):
        """Enter ] submode."""
        executor_sub = self.executor_sub_closesquarebracekt
        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def percent(self, num=1, num_str=""):
        """Find the next item in this line and execute."""
        motion_info = self.helper_motion.percent()
        return self.execute_func_deferred(motion_info)

    def f(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        executor_sub = self.executor_sub_f_t

        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def F(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        executor_sub = self.executor_sub_f_t

        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def t(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        executor_sub = self.executor_sub_f_t

        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def T(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        executor_sub = self.executor_sub_f_t

        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def semicolon(self, num=1, num_str=""):
        """Repeat latest f, t, F, T and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.semicolon(num)
        return self.execute_func_deferred(motion_info)

    def comma(self, num=1, num_str=""):
        """Repeat latest f, t, F, T in opposite direction and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.comma(num)
        return self.execute_func_deferred(motion_info)

    def apostrophe(self, num=1, num_str=""):
        """Motion to line of mark."""
        executor_sub = self.executor_sub_alnum
        self.set_parent_info_to_submode(executor_sub, num, num_str)

        def run(ch):
            motion_info = self.helper_motion.apostrophe(ch)
            return self.execute_func_deferred(motion_info)

        executor_sub.set_func_list_deferred([FUNC_INFO(run, True)])
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def backtick(self, num=1, num_str=""):
        """Motion to position of mark."""
        executor_sub = self.executor_sub_alnum
        self.set_parent_info_to_submode(executor_sub, num, num_str)

        def run(ch):
            motion_info = self.helper_motion.backtick(ch)
            return self.execute_func_deferred(motion_info)

        executor_sub.set_func_list_deferred([FUNC_INFO(run, True)])
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def i(self, num=1, num_str=""):
        """Select block exclusively."""
        executor_sub = self.executor_sub_motion_i

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def a(self, num=1, num_str=""):
        """Select block inclusively."""
        executor_sub = self.executor_sub_motion_a

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def slash(self, num=1, num_str=""):
        """Find text position."""
        executor_sub = self.executor_sub_search
        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, False)

    def n(self, num=1, num_str=""):
        """Goto the next selected text."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.n(num)
        return self.execute_func_deferred(motion_info)

    def N(self, num=1, num_str=""):
        """Goto the next selected text."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.N(num)
        return self.execute_func_deferred(motion_info)

    def space(self, num=1, num_str=""):
        """Go to [num] characters to the right and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.space(num)
        return self.execute_func_deferred(motion_info)

    def backspace(self, num=1, num_str=""):
        """Go to [num] characters to the left and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.backspace(num)
        return self.execute_func_deferred(motion_info)

    def enter(self, num=1, num_str=""):
        """Go to [num] lines downward linewise and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.enter(num)
        return self.execute_func_deferred(motion_info)

    def asterisk(self, num=1, num_str=""):
        """Search word under cursor forward and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.asterisk(num)
        return self.execute_func_deferred(motion_info)

    def sharp(self, num=1, num_str=""):
        """Search word under cursor backward and execute."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.sharp(num)
        return self.execute_func_deferred(motion_info)

    def run_easymotion(self, num=1, num_str=""):
        """Run easymotion."""
        executor_sub = self.executor_sub_easymotion

        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def z(self, num=1, num_str=""):
        """Execute sneak."""
        use_sneak = CONF.get(CONF_SECTION, "use_sneak")
        if use_sneak:
            executor_sub = self.executor_sub_sneak

            self.set_parent_info_to_submode(executor_sub, num, num_str)
            executor_sub.set_func_list_deferred(
                self.func_list_deferred, self.return_deferred
            )

            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)


class ExecutorSubMotion_d(ExecutorSubMotion):
    """Replace w, W for d command in normal."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.cmds += "s"
        cmds = "".join(re.escape(c) for c in self.cmds)
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))
        self.executor_sub_delete_surround = ExecutorDeleteSurround(vim_status)

    def w(self, num=1, num_str=""):
        """Move forward [num] words and delete."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.w_for_d(num)
        return self.execute_func_deferred(motion_info)

    def W(self, num=1, num_str=""):
        """Move forward [num] WORDs and delete."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.W_for_d(num)
        return self.execute_func_deferred(motion_info)

    def s(self, num=1, num_str=""):
        """Delete surroundings."""
        executor_sub = self.executor_sub_delete_surround
        self.set_parent_info_to_submode(executor_sub, num, num_str)
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)


class ExecutorSubMotion_c(ExecutorSubMotion):
    """Replace w, W for c command in normal."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.cmds += "s"
        cmds = "".join(re.escape(c) for c in self.cmds)
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))
        self.executor_sub_change_surround = ExecutorChangeSurround(vim_status)

    def w(self, num=1, num_str=""):
        """Move forward [num] words and delete and start insert mode."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.w_for_c(num)
        return self.execute_func_deferred(motion_info)

    def W(self, num=1, num_str=""):
        """Move forward [num] WORDs and delete and start insert mode."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.W_for_c(num)
        return self.execute_func_deferred(motion_info)

    def s(self, num=1, num_str=""):
        """Change surroundings."""
        executor_sub = self.executor_sub_change_surround
        self.set_parent_info_to_submode(executor_sub, num, num_str)
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)


class ExecutorSubMotion_y(ExecutorSubMotion_d):
    """Add s for surround to normal y command."""

    def __init__(self, vim_status):
        """Extend yank motions with surround support."""
        super().__init__(vim_status)
        self.cmds += "s"
        cmds = "".join(re.escape(c) for c in self.cmds)
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))

        self.executor_sub_motion_c = ExecutorSubMotion_c(vim_status)
        self.executor_sub_surround = ExecutorAddSurround(vim_status)

    def s(self, num=1, num_str=""):
        """Add surroundings."""
        executor_sub = self.executor_sub_motion_c
        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [
                FUNC_INFO(self.executor_sub_surround.set_motion_info, True),
            ]
        )
        executor_sub.return_deferred = RETURN_EXECUTOR_METHOD_INFO(
            self.executor_sub_surround, True
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)


class ExecutorSubSubCmd_g(ExecutorSubBase):
    """Submode of Submode of g."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

        cmds = "g"
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))

    def g(self, num=1, num_str=""):
        """Goto line (gg)."""
        num = self.parent_num[-1] * self.parent_num[0]
        if self.parent_num_str[-1] == "" and self.parent_num_str[0] == "":
            num = 1
        motion_info = self.helper_motion.G(num, True)
        return self.execute_func_deferred(motion_info)


class ExecutorSubCmd_g(ExecutorSubBase):
    """Submode of g."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

        cmds = "gdtTuUc~"
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))
        self.executor_sub_motion = ExecutorSubMotion(vim_status)

    def g(self, num=1, num_str=""):
        """Goto line(gg)."""
        num = self.parent_num[-1]
        if self.parent_num_str[-1] == "":
            num = 1
        motion_info = self.helper_motion.G(num, True)
        return self.execute_func_deferred(motion_info)

    def d(self, num=1, num_str=""):
        """Go to definition and update the jumplist."""
        vs = self.vim_status
        previous = vs.get_current_location()
        vs.push_jump()

        editor = self.get_editor()
        if editor:
            vs.start_definition_tracking(previous)
            editor.go_to_definition_from_cursor()

    def t(self, num=1, num_str=""):
        """Cycle to next file."""
        editor_stack = self.get_editorstack()
        if self.parent_num_str[-1] == "":
            editor_stack.tabs.tab_navigate(1)
        else:
            idx = self.parent_num[-1] - 1
            if editor_stack.get_stack_count() > idx:
                editor_stack.set_stack_index(idx)
        self.vim_status.set_focus_to_vim()

    def T(self, num=1, num_str=""):
        """Cycle to previous file."""
        editor_stack = self.get_editorstack()
        for _ in range(self.parent_num[-1]):
            editor_stack.tabs.tab_navigate(-1)
        self.vim_status.set_focus_to_vim()

    def u(self, num=1, num_str=""):
        """Make txt lowercase and move cursor to the start of selection."""
        if self.vim_status.is_normal():
            executor_sub = self.executor_sub_motion
            self.set_parent_info_to_submode(executor_sub, num, num_str)
            executor_sub.set_func_list_deferred(
                [FUNC_INFO(lambda x: self.helper_action.handle_case(x, "lower"), True)]
            )

            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)
        else:
            self.helper_action.handle_case(None, "lower")

    def U(self, num=1, num_str=""):
        """Make txt uppercase and move the cursor to the start of selection."""
        if self.vim_status.is_normal():
            executor_sub = self.executor_sub_motion
            self.set_parent_info_to_submode(executor_sub, num, num_str)
            executor_sub.set_func_list_deferred(
                [FUNC_INFO(lambda x: self.helper_action.handle_case(x, "upper"), True)]
            )

            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)
        else:
            self.helper_action.handle_case(None, "upper")

    def tilde(self, num=1, num_str=""):
        """Switch case of the character under the cursor.

        Move the cursor to the start of selection.
        """
        if self.vim_status.is_normal():
            executor_sub = self.executor_sub_motion
            self.set_parent_info_to_submode(executor_sub, num, num_str)
            executor_sub.set_func_list_deferred(
                [FUNC_INFO(lambda x: self.helper_action.handle_case(x, "swap"), True)]
            )

            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)
        else:
            self.helper_action.handle_case(None, "swap")

    def c(self, num=1, num_str=""):
        """Toggle comment."""
        if self.vim_status.is_normal():
            executor_sub = self.executor_sub_motion
            self.set_parent_info_to_submode(executor_sub, num, num_str)
            executor_sub.set_func_list_deferred(
                [FUNC_INFO(lambda x: self.helper_action.toggle_comment(x), True)]
            )

            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)
        else:
            self.helper_action.toggle_comment(None)


class ExecutorSubCmd_register(ExecutorSubBase):
    """Set register name."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

    def __call__(self, ch: str):
        """Update command to vim status."""
        self.update_input_cmd_info(None, None, ch)

        self.vim_status.sub_mode = None

        return True


class ExecutorSubCmd_f_t(ExecutorSubBase):
    """Submode of f, t."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

    def __call__(self, ch: str):
        """Go to the occurrence of character and execute."""
        ch_previous = self.vim_status.input_cmd.cmd[-1]
        method = self.helper_motion.find_cmd_map.get(ch_previous, None)

        self.update_input_cmd_info(None, None, ch)

        self.vim_status.sub_mode = None

        if method:
            if len(self.parent_num) == 1:
                num = self.parent_num[0]
            else:
                num = self.parent_num[0] * self.parent_num[-1]
            motion_info = method(ch, num)
            return self.process_return(self.execute_func_deferred(motion_info))

        return True


class ExecutorSubCmdSneak(ExecutorSubBase):
    """Submode of sneak"""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

    def __call__(self, ch2: str):
        """Go to the occurrence of character and execute."""
        if len(ch2) < 2:
            return False
        ch_previous = self.vim_status.input_cmd.cmd[-1]
        method = self.helper_motion.find_cmd_map.get(ch_previous, None)

        self.update_input_cmd_info(None, None, ch2)

        self.vim_status.sub_mode = None

        if method:
            if len(self.parent_num) == 1:
                num = self.parent_num[0]
            else:
                num = self.parent_num[0] * self.parent_num[-1]
            motion_info = method(ch2, num)
            return self.process_return(self.execute_func_deferred(motion_info))

        return True


class ExecutorSubCmd_r(ExecutorSubBase):
    """Submode of r."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

        self.pos_start = 0
        self.pos_end = 0

    def __call__(self, ch: str):
        """Replace the character under the cursor with ch."""
        self.update_input_cmd_info(None, None, ch)

        self.helper_action.replace_txt_with_ch(self.pos_start, self.pos_end, ch)

        self.vim_status.sub_mode = None

        self.execute_func_deferred(None)

        return True


class ExecutorSubCmd_Z(ExecutorSubBase):
    """Submode of Z."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

        cmds = "ZQ"
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))
        self.editor_widget = vim_status.editor_widget

    def Q(self, num=1, num_str=""):
        """Close current file without saving."""
        # TODO :
        self.close_current_file()
        self.vim_status.set_focus_to_vim()

    def Z(self, num=1, num_str=""):
        """Save and close current file."""
        self.save_current_file()
        self.close_current_file()
        self.vim_status.set_focus_to_vim()


class ExecutorSearch(ExecutorSubBase):
    """Submode of search."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

    def __call__(self, txt: str) -> bool:
        """Parse txt and executor command.

        Returns:
            if return is True, Clear command line

        """
        if txt[-1] == "\b":
            cmd_line = self.vim_status.cmd_line
            if len(txt) <= 2:
                self.vim_status.sub_mode = None
                return True
            else:
                cmd_line.setText(txt[:-2])
        if txt[-1] != "\r":
            return False

        if txt[0] != "/" or len(txt) <= 2:
            self.vim_status.sub_mode = None
            return True

        # '/' is saved to input cmd_info at ExecutorBase.
        # So we need only txt[1:].
        self.update_input_cmd_info(None, None, txt[1:])

        txt = txt[1:-1]  # remove /, \r

        self.helper_motion.search(txt)
        motion_info = self.helper_motion.n(1, "")
        self.vim_status.sub_mode = None

        self.vim_status.push_jump()
        ret = self.process_return(self.execute_func_deferred(motion_info))
        self.vim_status.push_jump()
        return ret


class ExecutorSubCmd_alnum(ExecutorSubBase):
    """Allow the alphabetics and numbers as input."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

    def __call__(self, ch: str):
        """Return deferred result when ``ch`` is alphanumeric."""
        self.update_input_cmd_info(None, None, ch)

        self.vim_status.sub_mode = None

        if ch.isalnum():
            return self.process_return(self.execute_func_deferred(ch))

        return True


class ExecutorSubCmd_opensquarebracket(ExecutorSubBase):
    """submode of ["""

    def __init__(self, vim_status):
        """Initialize submode for ``[`` commands."""
        super().__init__(vim_status)
        self.allow_leaderkey = False

        self.has_zero_cmd = False

        cmds = "".join(re.escape(c) for c in "dc[")
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))

    def d(self, num=1, num_str=""):
        """Go to previous warning/error."""
        num = num * self.parent_num[0]
        editor = self.get_editor()
        for _ in range(num):
            editor.go_to_previous_warning()
        self.vim_status.cursor.draw_vim_cursor()

    def opensquarebracket(self, num=1, num_str=""):
        """Jump to previous Python definition."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.prev_pydef(num)
        return self.execute_func_deferred(motion_info)

    def c(self, num=1, num_str=""):
        """Move to the previous cell header."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.prev_cell(num)
        return self.execute_func_deferred(motion_info)


class ExecutorSubCmd_closesquarebracket(ExecutorSubBase):
    """submode of ]"""

    def __init__(self, vim_status):
        """Initialize submode for ``]`` commands."""
        super().__init__(vim_status)
        self.allow_leaderkey = False

        self.has_zero_cmd = False

        cmds = "".join(re.escape(c) for c in "dc]")
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))

    def d(self, num=1, num_str=""):
        """Go to next warning/error."""
        num = num * self.parent_num[0]
        editor = self.get_editor()
        for _ in range(num):
            editor.go_to_next_warning()
        self.vim_status.cursor.draw_vim_cursor()

    def closesquarebracket(self, num=1, num_str=""):
        """Jump to next Python definition."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.next_pydef(num)
        return self.execute_func_deferred(motion_info)

    def c(self, num=1, num_str=""):
        """Move to the next cell header."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.next_cell(num)
        return self.execute_func_deferred(motion_info)


class ExecutorSubCmd_openbrace(ExecutorSubBase):
    """Submode of {"""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

        self.has_zero_cmd = False

        cmds = "{"
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))

    def openbrace(self, num=1, num_str=""):
        """Jump to previous Python block."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.prev_pyblock(num)
        return self.execute_func_deferred(motion_info)


class ExecutorSubCmd_closebrace(ExecutorSubBase):
    """Submode of }"""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

        self.has_zero_cmd = False

        cmds = "}"
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))

    def closebrace(self, num=1, num_str=""):
        """Jump to next Python block."""
        num = num * self.parent_num[0]
        motion_info = self.helper_motion.next_pyblock(num)
        return self.execute_func_deferred(motion_info)


class ExecutorSubCmd_z(ExecutorSubBase):
    """Submode of z."""

    def __init__(self, vim_status):
        """Initialize submode for ``z`` commands."""
        super().__init__(vim_status)
        self.allow_leaderkey = False

        self.has_zero_cmd = False

        cmds = "ztb"
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))

    def scroll_cursor_to_center(self, offset: int):
        """Scroll cursor line to center."""
        pos = self.vim_status.get_cursor().position()

        editor = self.get_editor()
        line, _ = editor.get_cursor_line_column()
        editor.go_to_line(line + offset)

        self.vim_status.cursor.set_cursor_pos(pos)
        self.vim_status.cursor.draw_vim_cursor()

    def t(self, num=1, num_str=""):
        """Cursor line to top of screen."""
        visible_lines = self.vim_status.get_number_of_visible_lines()
        offset = visible_lines // 2 - 1
        self.scroll_cursor_to_center(offset)

        editor = self.get_editor()
        n_block = editor.blockCount()
        idx_block, _ = editor.get_cursor_line_column()
        leftover = offset - (n_block - idx_block)

        # When current line is near the end of the document, vim control scrollbar.
        if leftover > 0:
            scroll = editor.verticalScrollBar()
            v_max = scroll.maximum()
            v_current = scroll.value()
            offset_scroll = min([v_max, v_current + leftover])
            scroll.setValue(offset_scroll)

    def b(self, num=1, num_str=""):
        """Cursor line to bottom of screen."""
        visible_lines = self.vim_status.get_number_of_visible_lines()
        self.scroll_cursor_to_center(-(visible_lines // 2 - 1))

    def z(self, num=1, num_str=""):
        """Cursor line to center of screen."""
        self.scroll_cursor_to_center(1)



# Third Party Libraries
import pytest
from qtpy.QtCore import Qt, QEvent
from qtpy.QtGui import QKeyEvent, QStandardItem
from qtpy.QtWidgets import QDialog

# Project Libraries
from spyder_okvim.utils.list_dialog import PopupListDialog, PopupTableDialog
from spyder_okvim.utils.jump_dialog import JumpListDialog
from spyder_okvim.utils.mark_dialog import MarkListDialog
from spyder_okvim.utils.jump_list import JumpList


class DummyVimStatus:
    """Simple stand-in for VimStatus used in dialogs."""

    def __init__(self, jump_list):
        self.jump_list = jump_list
        self.focused = False

    def jump_backward(self):
        self.jump_list.back()

    def jump_forward(self):
        self.jump_list.forward()

    def set_focus_to_vim(self):
        self.focused = True


def test_popup_list_dialog_navigation(qtbot):
    dlg = PopupListDialog("Test", min_width=100)
    qtbot.addWidget(dlg)
    dlg.list_model.setStringList(["a", "b", "c"])
    dlg.list_viewer.setCurrentIndex(dlg.list_model.index(0, 0))
    dlg.show()
    qtbot.waitExposed(dlg)

    dlg.next_row()
    assert dlg.list_viewer.currentIndex().row() == 1
    dlg.prev_row()
    assert dlg.list_viewer.currentIndex().row() == 0

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_N, Qt.ControlModifier)
    dlg.keyPressEvent(event)
    assert dlg.list_viewer.currentIndex().row() == 1
    event = QKeyEvent(QEvent.KeyPress, Qt.Key_P, Qt.ControlModifier)
    dlg.keyPressEvent(event)
    assert dlg.list_viewer.currentIndex().row() == 0

    dlg.keyPressEvent(QKeyEvent(QEvent.KeyPress, Qt.Key_Return, Qt.NoModifier))
    assert dlg.result() == QDialog.Accepted


def test_popup_list_dialog_escape_and_pages(qtbot):
    dlg = PopupListDialog("Test")
    qtbot.addWidget(dlg)
    dlg.list_model.setStringList(["a", "b", "c", "d"])
    dlg.list_viewer.setCurrentIndex(dlg.list_model.index(0, 0))
    dlg.show()
    qtbot.waitExposed(dlg)

    dlg.pg_down()
    dlg.pg_up()
    dlg.pg_half_down()
    dlg.pg_half_up()
    dlg.get_number_of_visible_lines()

    dlg.keyPressEvent(QKeyEvent(QEvent.KeyPress, Qt.Key_Escape, Qt.NoModifier))
    assert dlg.result() == QDialog.Rejected


def test_popup_table_dialog_keys(qtbot):
    dlg = PopupTableDialog("Table", headers=["col"])
    qtbot.addWidget(dlg)
    dlg.list_model.appendRow([QStandardItem("x")])
    dlg.list_model.appendRow([QStandardItem("y")])
    dlg.list_viewer.setCurrentIndex(dlg.list_model.index(0, 0))
    dlg.list_viewer.selectRow(0)
    dlg.show()
    qtbot.waitExposed(dlg)

    dlg.keyPressEvent(QKeyEvent(QEvent.KeyPress, Qt.Key_N, Qt.ControlModifier))
    assert dlg.list_viewer.currentIndex().row() == 1

    dlg.keyPressEvent(QKeyEvent(QEvent.KeyPress, Qt.Key_Return, Qt.NoModifier))
    assert dlg.result() == QDialog.Accepted


def test_popup_table_dialog_escape(qtbot):
    dlg = PopupTableDialog("Table", headers=["col"], min_width=120)
    qtbot.addWidget(dlg)
    dlg.list_model.appendRow([QStandardItem("x")])
    dlg.list_viewer.setCurrentIndex(dlg.list_model.index(0, 0))
    dlg.list_viewer.selectRow(0)
    dlg.show()
    qtbot.waitExposed(dlg)

    dlg.pg_up()
    dlg.pg_down()
    dlg.pg_half_up()
    dlg.pg_half_down()
    dlg.get_number_of_visible_lines()

    dlg.keyPressEvent(QKeyEvent(QEvent.KeyPress, Qt.Key_Escape, Qt.NoModifier))
    assert dlg.result() == QDialog.Rejected


def test_jump_list_dialog_navigation(qtbot, tmpdir):
    file1 = tmpdir.join("a.txt")
    file1.write("alpha\nbeta\n")
    file2 = tmpdir.join("b.txt")
    file2.write("gamma\n")

    jl = JumpList()
    jl.push(str(file1), 0)
    jl.push(str(file2), 0)
    vs = DummyVimStatus(jl)

    dlg = JumpListDialog(vs)
    qtbot.addWidget(dlg)
    dlg.show()
    qtbot.waitExposed(dlg)

    assert dlg.list_model.rowCount() == 2
    assert dlg.list_model.item(1, 0).text().startswith(">")

    qtbot.keyPress(dlg.list_viewer, Qt.Key_O, modifier=Qt.ControlModifier)
    assert vs.jump_list.index == 1

    qtbot.keyPress(dlg.list_viewer, Qt.Key_I, modifier=Qt.ControlModifier)
    assert vs.jump_list.index == 2

    idx_before = vs.jump_list.index
    qtbot.keyPress(dlg.list_viewer, Qt.Key_N, modifier=Qt.ControlModifier)
    assert vs.jump_list.index == idx_before

    dlg.reject()
    assert vs.focused


def test_mark_list_dialog_accept_reject(qtbot, tmpdir):
    file1 = tmpdir.join("c.txt")
    file1.write("line1\nline2\n")
    marks = [("a", {"file": str(file1), "line": 1, "col": 0})]

    dlg = MarkListDialog(marks)
    qtbot.addWidget(dlg)
    dlg.show()
    qtbot.waitExposed(dlg)

    assert dlg.list_model.rowCount() == 1
    row = dlg.list_model.item(0, 0).text()
    assert row == "a"

    dlg.accept()
    assert dlg.get_selected_mark() == "a"

    dlg.reject()
    assert dlg.get_selected_mark() == ""


def test_jump_list_get_line_info(tmpdir):
    file1 = tmpdir.join("eof.txt")
    file1.write("foo")
    jl = JumpList()
    vs = DummyVimStatus(jl)
    dlg = JumpListDialog(vs)
    line, col, text = dlg._get_line_info(str(file1), 1)
    assert (line, col, text) == (1, 2, "foo")

    missing = tmpdir.join("missing.txt")
    line, col, text = dlg._get_line_info(str(missing), 0)
    assert (line, col, text) == (1, 1, "")


def test_mark_list_dialog_invalid_accept_and_populate(tmpdir):
    bad_file = tmpdir.join("missing.txt")
    marks = [("a", {"file": str(bad_file), "line": 0, "col": 0})]
    dlg = MarkListDialog(marks)
    dlg.list_viewer.setCurrentIndex(dlg.list_model.index(1, 0))
    dlg.accept()
    assert dlg.get_selected_mark() == ""
    assert dlg.list_model.item(0, 4).text() == ""



# -*- coding: utf-8 -*-
"""Executor handling leader-key sequences."""

# Third Party Libraries
from qtpy.QtCore import QTimer
from spyder.api.plugins import Plugins

# Project Libraries
from spyder_okvim.executor.executor_base import ExecutorBase


class ExecutorLeaderKey(ExecutorBase):
    """Executor for leader key."""

    def __init__(self, vim_status):
        super().__init__(vim_status)

        self.dispatcher = {
            "i": self.auto_import,
            "b": self.toggle_breakpoint,
            "\r": self.run_cell_and_advance,
            "r": self.run_selection,
            "f": self.formatting,
            "p": self.open_switcher,
            "s": self.open_symbol_swicher,
        }

        self.set_selection_to_editor_using_vim_selection = (
            self.vim_status.cursor.set_selection_to_editor_using_vim_selection
        )
        self.prev_executor = None

    def set_easymotion_key(self, key):
        "Set key for easymotion."
        self.dispatcher[key] = self.execute_easymotion

    def __call__(self, txt):
        """Dispatch leader-key command ``txt`` and return execution info."""
        method = self.dispatcher.get(txt, None)

        ret = None
        if method:
            ret = method(1, "")

        if ret:
            self.vim_status.sub_mode = ret.sub_mode
            return ret.clear_command_line
        else:
            self.vim_status.sub_mode = None
            return True

    def auto_import(self, num=1, num_str=""):
        """Insert import statements from user defined list(for personal)."""
        editor = self.get_editor()
        try:
            editor.auto_import.auto_import()
            self.vim_status.set_focus_to_vim()
        except AttributeError:
            pass

    def toggle_breakpoint(self, num=1, num_str=""):
        """Toggle break."""
        editor = self.get_editor()
        editor.breakpoints_manager.toogle_breakpoint()

    def run_cell_and_advance(self, num=1, num_str=""):
        """Run cell and advance."""
        editor_stack = self.get_editorstack()
        editor_stack.sig_trigger_action.emit("run cell and advance", Plugins.Run)
        self.vim_status.set_focus_to_vim()

    def run_selection(self, num=1, num_str=""):
        """Run selected text or current line in console."""
        editor_stack = self.get_editorstack()
        editor = self.get_editor()
        old_cursor = self.set_selection_to_editor_using_vim_selection()
        editor_stack.sig_trigger_action.emit("run selection and advance", Plugins.Run)
        if old_cursor:
            editor.setTextCursor(old_cursor)
        self.vim_status.set_focus_to_vim()

    def retrieve_curosr_pos(self, pos: int, delay: int):
        """Retrieve the cursor position."""

        def _retrieve():
            self.vim_status.cursor.set_cursor_pos(pos)
            self.vim_status.set_focus_to_vim()
            self.vim_status.cursor.draw_vim_cursor()

        QTimer.singleShot(delay, _retrieve)

    def formatting(self, num=1, num_str=""):
        """Format document automatically."""
        editor = self.get_editor()
        editor.format_document_or_range()

    def open_switcher(self, num=1, num_str=""):
        """Open switcher for buffers."""
        main = self.vim_status.main
        switcher = main.get_plugin(Plugins.Switcher)
        switcher.open_switcher()

    def open_symbol_swicher(self, num=1, num_str=""):
        """Open switcher for symbol."""
        main = self.vim_status.main
        switcher = main.get_plugin(Plugins.Switcher)
        switcher.open_switcher(symbol=True)

    def execute_easymotion(self, num=1, num_str=""):
        """Execute easymotion."""
        return self.prev_executor.run_easymotion()



# -*- coding: utf-8 -*-
"""Expose all available command executors.

This package defines the main entry points used by OkVim to interpret
keystrokes.  Normal, visual and vline executors emulate Vim behaviour in
the Spyder editor, while ``ExecutorLeaderKey`` handles sequences that
start with the configurable leader key.
"""
# %% Import
# Local imports
from .executor_normal import ExecutorNormalCmd
from .executor_visual import ExecutorVisualCmd
from .executor_vline import ExecutorVlineCmd
from .executor_leader import ExecutorLeaderKey



# -*- coding: utf-8 -*-
"""Core data structures for Vim status handling."""

from qtpy.QtCore import QEvent
from qtpy.QtGui import QKeyEvent


class VimState:
    """Vim state constants."""

    VISUAL = 0
    VLINE = 1
    NORMAL = 3
    INSERT = 4
    SUBMODE = 5


class FindInfo:
    """Metadata for ``f``, ``F``, ``t`` and ``T`` commands."""

    def __init__(self):
        self.cmd_name: str = ""
        self.ch: str = ""

    def set(self, cmd_name: str, ch: str):
        """Update this command information.

        Args:
            cmd_name: Command name (``f``/``F``/``t``/``T``).
            ch: Character used with the command.
        """
        self.cmd_name = cmd_name
        self.ch = ch


class InputCmdInfo:
    """Container for the command currently being typed."""

    def __init__(self, num_str: str, cmd: str):
        """Create a new instance.

        Args:
            num_str: Number prefix for the command.
            cmd: Command string.
        """
        self.num_str: str = num_str
        self.cmd: str = cmd

    def clear(self):
        """Reset both number and command strings."""
        self.num_str = ""
        self.cmd = ""

    def set(self, input_cmd_info):
        """Copy the fields from another ``InputCmdInfo``.

        Args:
            input_cmd_info: Source command info to copy from.
        """
        self.num_str = input_cmd_info.num_str
        self.cmd = input_cmd_info.cmd


class DotCmdInfo:
    """Information about the last ``.`` command."""

    def __init__(self):
        self.vim_state = None
        self.sel_block_no_start = None
        self.sel_block_no_end = None
        self.sel_col_start = None
        self.sel_col_end = None
        self.num_str: str = ""
        self.cmd: str = ""
        self.cmd_list_insertmode = None
        self.register_name = None
        self.editor_connected = None
        self.key_list_from_editor = []
        self.key_list_to_cmd_line = []

    def clear_key_list(self):
        """Remove any stored key events."""
        self.key_list_from_editor.clear()
        self.key_list_to_cmd_line.clear()

    def to_cmd_string(self, num, num_str):
        """Return the textual representation of this dot command.

        Args:
            num: Numeric prefix used for the command.
            num_str: Original numeric string typed by the user.
        """
        # TODO : apply register name.
        cmd_str = ""

        if self.vim_state == VimState.VLINE:
            cmd_str += "V"
            n_lines = self.sel_block_no_end - self.sel_block_no_start
            if n_lines > 0:
                cmd_str += f"{n_lines}j"
            cmd_str += self.cmd
            return cmd_str

        if self.vim_state == VimState.VISUAL:
            cmd_str += "v"
            n_lines = self.sel_block_no_end - self.sel_block_no_start
            if n_lines > 0:
                cmd_str += f"{n_lines}j0"

                if self.sel_col_end > 1:
                    cmd_str += f"{self.sel_col_end - 1}l"
            else:
                n = self.sel_col_end - self.sel_col_start - 1
                if n > 0:
                    cmd_str += f"{n}l"
            cmd_str += self.cmd
            return cmd_str

        if num_str:
            return str(num) + self.cmd
        else:
            return self.num_str + self.cmd



class KeyInfo:
    """Serializable representation of a :class:`QKeyEvent`."""

    def __init__(self, key_code: int, text: str, modifiers: int, identifier: int = 1):
        self.key_code = key_code
        self.text = text
        self.modifiers = modifiers
        #: 0 if coming from the Vim command line, 1 if from the editor
        self.identifier = identifier

    def to_event(self):
        """Return a :class:`QKeyEvent` created from this info."""
        event = QKeyEvent(QEvent.KeyPress, self.key_code, self.modifiers, self.text)
        event.ignore()
        return event


class RegisterInfo:
    """Simple register structure used by macros and yank/put."""

    def __init__(self):
        """Initialize an empty register."""
        self.name = ""
        self.content = ""
        self.type = VimState.NORMAL





from __future__ import annotations

# Standard Libraries
import os.path as osp

# Third Party Libraries
from qtpy.QtCore import QEvent, Qt
from qtpy.QtGui import QStandardItem

from .list_dialog import PopupTableDialog


class JumpListDialog(PopupTableDialog):
    """Dialog to display the jump list."""

    _MIN_WIDTH = 800
    _MAX_HEIGHT = 600

    def __init__(self, vim_status, parent=None) -> None:
        super().__init__(
            "Jumps",
            parent=parent,
            headers=["#", "Line", "Col", "File", "Text"],
            min_width=self._MIN_WIDTH,
            max_height=self._MAX_HEIGHT,
        )

        self.vim_status = vim_status
        self.jump_list = vim_status.jump_list

        self._populate()
        self.update_current_row()

        # Forward key events from the list viewer to this dialog
        self.list_viewer.installEventFilter(self)

        # When closed return focus to command line
        self.finished.connect(lambda *_: self.vim_status.set_focus_to_vim())

    # ------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------
    def _get_line_info(self, file_path: str, pos: int) -> tuple[int, int, str]:
        """Return line, column and text for *pos* in *file_path*."""
        line = col = 0
        text = ""
        try:
            with open(file_path, "r", encoding="utf-8") as fh:
                data = fh.read()
            line = data.count("\n", 0, pos)
            line_start = data.rfind("\n", 0, pos) + 1
            line_end = data.find("\n", pos)
            if line_end == -1:
                line_end = len(data)
            col = pos - line_start
            text = data[line_start:line_end].strip()
        except Exception:
            pass
        return line + 1, col + 1, text

    def _populate(self) -> None:
        self.list_model.setRowCount(0)
        for i, jump in enumerate(self.jump_list.jumps, start=1):
            line, col, text = self._get_line_info(jump.file, jump.pos)
            basename = osp.basename(jump.file)
            mark = ">" if i == self.jump_list.index else ""
            row = [
                QStandardItem(f"{mark}{i}"),
                QStandardItem(str(line)),
                QStandardItem(str(col)),
                QStandardItem(basename),
                QStandardItem(text),
            ]
            for idx, item in enumerate(row):
                item.setEditable(False)
                if idx in (1, 2):
                    item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
            self.list_model.appendRow(row)

    def update_current_row(self) -> None:
        row = max(0, self.jump_list.index - 1)
        if self.list_model.rowCount() > 0:
            self.list_viewer.setCurrentIndex(self.list_model.index(row, 0))
            self.list_viewer.selectRow(row)

    # ------------------------------------------------------------------
    # Qt overrides
    # ------------------------------------------------------------------
    def keyPressEvent(self, event) -> None:
        mod = event.modifiers()
        key = event.key()
        if mod == Qt.ControlModifier and key == Qt.Key_O:
            self.vim_status.jump_backward()
            self._populate()
            self.update_current_row()
            return
        if mod == Qt.ControlModifier and key == Qt.Key_I:
            self.vim_status.jump_forward()
            self._populate()
            self.update_current_row()
            return
        if mod == Qt.ControlModifier and key in (Qt.Key_N, Qt.Key_P):
            # Prevent `Ctrl+n`/`Ctrl+p` from navigating the list
            # (these shortcuts are used in other dialogs)
            return
        super().keyPressEvent(event)

    def eventFilter(self, obj, event):
        """Handle shortcut events forwarded from the list viewer."""
        if obj is self.list_viewer and event.type() in (
            QEvent.ShortcutOverride,
            QEvent.KeyPress,
        ):
            self.keyPressEvent(event)
            return True
        return super().eventFilter(obj, event)



"""Decorators for managing submode execution in the Spyder OkVim executor framework.

This module provides decorators to help with entering and managing submodes
within the executor system. The main decorator, `submode`, wraps command methods
to prepare submode executors, attach deferred callbacks, and ensure a clean
context switch using RETURN_EXECUTOR_METHOD_INFO.

Typical usage involves applying the `submode` decorator to executor methods
that need to transition into a submode, optionally providing a function to
retrieve a list of deferred functions for the submode.
"""

# Standard Libraries
from collections.abc import Callable
from functools import wraps

from .executor_base import FUNC_INFO, RETURN_EXECUTOR_METHOD_INFO


def submode(
    func_list_getter: Callable[[object], list[FUNC_INFO]] | None = None,
    clear_command_line: bool = True,
):
    """Return a decorator that prepares and enters a submode.

    The wrapped command should return the submode executor.  This helper
    attaches any deferred callbacks defined by ``func_list_getter`` and
    finally returns a :class:`RETURN_EXECUTOR_METHOD_INFO` instance so the
    caller can switch context cleanly.
    """

    def decorator(func: Callable):
        @wraps(func)
        def wrapper(self, num=1, num_str=""):
            executor_sub = func(self, num, num_str)
            self.set_parent_info_to_submode(executor_sub, num, num_str)
            func_list = func_list_getter(self) if func_list_getter else []
            if not executor_sub.func_list_deferred:
                executor_sub.set_func_list_deferred(func_list)
            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, clear_command_line)

        return wrapper

    return decorator



# -*- coding: utf-8 -*-
"""Main object storing global Vim state."""

# Standard Libraries
import os.path as osp
from collections import defaultdict

# Third Party Libraries
from qtpy.QtCore import QCoreApplication, QObject, QTimer, Signal, Slot
from qtpy.QtGui import QKeyEvent, QTextCursor
from qtpy.QtWidgets import QApplication, QLabel, QWidget
from spyder.api.plugins import Plugins
from spyder.config.manager import CONF

# Project Libraries
from spyder_okvim.spyder.config import CONF_SECTION
from spyder_okvim.utils.bookmark_manager import BookmarkManager
from spyder_okvim.utils.easymotion import EasyMotionMarkerManager, EasyMotionPainter
from spyder_okvim.utils.jump_list import JumpList
from spyder_okvim.utils.cell_helpers import CellRegion, get_document_cells

from .cursor import VimCursor
from .label import InlineLabel
from .macro import MacroManager
from .search import SearchInfo
from .state import DotCmdInfo, FindInfo, InputCmdInfo, KeyInfo, RegisterInfo, VimState


class VimStatus(QObject):
    """Track the global Vim emulation state."""

    change_label = Signal(int)

    def __init__(
        self, editor_widget: QWidget, main: QWidget, msg_label: QLabel
    ) -> None:
        """Initialize the status object.

        Args:
            editor_widget: Editor plugin used to access the current editor.
            main: Main Spyder window.
            msg_label: Label widget used to display status messages.
        """
        super().__init__()
        self.is_visual_mode = False
        self.vim_state = VimState.NORMAL
        self.editor_widget = editor_widget
        self.cursor: VimCursor = VimCursor(editor_widget)
        self.main = main

        # method mapping
        self.set_cursor = self.cursor.set_cursor
        self.get_cursor = self.cursor.get_cursor
        self.get_editor = self.cursor.get_editor
        self.get_editorstack = self.cursor.get_editorstack
        self.get_block_no_start_in_selection = (
            self.cursor.get_block_no_start_in_selection
        )
        self.get_block_no_end_in_selection = self.cursor.get_block_no_end_in_selection
        self.get_pos_start_in_selection = self.cursor.get_pos_start_in_selection
        self.get_pos_end_in_selection = self.cursor.get_pos_end_in_selection

        self.cmd_line = None

        self.sub_mode = None

        # command
        self.find_info = FindInfo()
        self.input_cmd = InputCmdInfo("", "")
        self.input_cmd_prev = InputCmdInfo("", "")
        self.dot_cmd = DotCmdInfo()
        self.running_dot_cmd = False

        # register
        self.register_dict = defaultdict(RegisterInfo)

        # config
        self.indent = "    "

        # search
        self.search = SearchInfo(self.cursor)

        # message
        self.msg_label = msg_label
        self.msg_prefix = ""

        # Macro
        self.manager_macro = MacroManager()

        # bookmarks
        bookmarks_file = osp.join(
            CONF.get_plugin_config_path(CONF_SECTION), "bookmarks.json"
        )
        self._application_plugin = None
        self.bookmark_manager = BookmarkManager(
            bookmarks_file,
            self._open_file_in_application,
            self.get_editorstack,
            self.get_editor,
            self.cursor.set_cursor_pos,
        )
        # Expose dictionaries for backwards compatibility
        self.bookmarks = self.bookmark_manager.bookmarks
        self.bookmarks_global = self.bookmark_manager.bookmarks_global

        # jumplist
        self.jump_list = JumpList()
        self.timer_go_to_definition = None

        # easymotion
        self.painter_easymotion = EasyMotionPainter()
        self.manager_marker_easymotion = EasyMotionMarkerManager()
        self.editor_connected_easymotion = None

        # Sneak
        self.n_annotate_max = 50
        self.labels_for_annotate = [
            InlineLabel(self.main) for _ in range(self.n_annotate_max)
        ]
        self.hide_annotate_on_txt()

    def clear_state(self):
        """Clear."""
        self.is_visual_mode = False
        self.vim_state = VimState.NORMAL
        editor = self.get_editor()
        if editor:
            editor.clear_extra_selections("vim_selection")
            editor.clear_extra_selections("vim_cursor")

    def is_normal(self):
        """Check that vim state is normal mode."""
        return self.vim_state == VimState.NORMAL

    def to_normal(self):
        """Change vim state to normal mode."""
        self.clear_state()
        cursor = self.get_cursor()
        if cursor.atBlockEnd() and not cursor.atBlockStart():
            pos = cursor.position()
            self.cursor.set_cursor_pos(pos - 1)
        try:
            self.change_label.emit(VimState.NORMAL)
        except RuntimeError:
            pass

    def get_cells(self) -> list[CellRegion]:
        """Return the cell regions for the active editor."""
        editor = self.get_editor()
        return get_document_cells(editor)

    def reset_for_test(self):
        """Reset status for test."""
        # Remove any pending EasyMotion overlays before processing Qt events.
        self.remove_marker_of_easymotion()
        QCoreApplication.processEvents()
        self.clear_state()

        self.find_info = FindInfo()
        self.input_cmd = InputCmdInfo("", "")
        self.input_cmd_prev = InputCmdInfo("", "")
        self.dot_cmd.clear_key_list()
        self.running_dot_cmd = False

        self.get_editor().clear_extra_selections("hl_yank")

        # register
        self.register_dict = defaultdict(RegisterInfo)

        # config
        self.indent = "    "

        # search
        self.search = SearchInfo(self.cursor)

        # Macro
        self.manager_macro = MacroManager()

        self.msg_label.setText("")

        # bookmarks
        self.bookmark_manager.clear()
        self.bookmarks = self.bookmark_manager.bookmarks
        self.bookmarks_global = self.bookmark_manager.bookmarks_global

        # jumplist
        self.jump_list = JumpList()

        # Ensure EasyMotion overlays are removed between tests to avoid
        # accessing deleted widgets when reusing the session-scoped editor.
        self.remove_marker_of_easymotion()

        if self.timer_go_to_definition is not None:
            self.timer_go_to_definition.stop()
            self.timer_go_to_definition = None

        self.to_normal()

    def to_insert(self):
        """Change vim state to insert mode."""
        editor = self.get_editor()
        if editor is None:
            return
        self.is_visual_mode = False
        self.remove_marker_of_easymotion()
        editor.clear_extra_selections("vim_selection")
        editor.clear_extra_selections("vim_cursor")
        self.change_label.emit(VimState.INSERT)

    def to_visual_char(self):
        """Change vim state to visual mode."""
        self.clear_state()
        self.is_visual_mode = True
        self.vim_state = VimState.VISUAL

        pos = self.get_cursor().position()
        self.cursor.create_selection(pos, pos + 1)

        self.change_label.emit(VimState.VISUAL)

    def to_visual_line(self):
        """Change vim state to vline mode."""
        self.clear_state()
        self.is_visual_mode = True
        self.vim_state = VimState.VLINE

        cursor = self.get_cursor()
        cursor.movePosition(QTextCursor.StartOfBlock)

        pos_start = cursor.position()

        cursor.movePosition(QTextCursor.EndOfBlock)
        pos_end = cursor.position()

        self.cursor.create_selection(pos_start, pos_end)

        self.change_label.emit(VimState.VLINE)

    def get_number_of_visible_lines(self):
        """Get the number of visible lines in editor."""
        editor = self.get_editor()
        num_lines = editor.viewport().height() // editor.fontMetrics().height()
        return num_lines

    def set_focus_to_vim(self):
        """Set focus to vim command line."""
        if self.cmd_line:
            self.cmd_line.setFocus()

    def set_focus_to_vim_after_delay(self, delay=300):
        """Set focus to the Vim command line after ``delay`` milliseconds.

        Note: This is mainly used when the editor regains focus after external
        actions.

        Args:
            delay: Delay in milliseconds before focusing the command line.
        """

        def focus():
            self.set_focus_to_vim()
            self.cursor.draw_vim_cursor()

        if self.cmd_line:
            QTimer.singleShot(delay, focus)

    @Slot(QKeyEvent)
    def rcv_key_from_editor(self, event):
        """Add key event from editor to list."""
        self.dot_cmd.key_list_from_editor.append(
            KeyInfo(event.key(), event.text(), event.modifiers(), 1)
        )

    def disconnect_from_editor(self):
        """Disconnect from the editor."""
        # disconnect previous connection.
        editor = self.dot_cmd.editor_connected
        if editor:
            editor.sig_key_pressed.disconnect(self.rcv_key_from_editor)
        self.dot_cmd.editor_connected = None

    def update_dot_cmd(
        self, connect_editor, register_name=None, key_list_to_cmd_line=None
    ):
        """Update input command info to dot command info."""
        if self.running_dot_cmd is True:
            return
        dot_cmd = self.dot_cmd
        dot_cmd.vim_state = self.vim_state
        dot_cmd.num_str = self.input_cmd.num_str
        dot_cmd.cmd = self.input_cmd.cmd
        dot_cmd.register_name = register_name
        dot_cmd.clear_key_list()
        if key_list_to_cmd_line:
            dot_cmd.key_list_to_cmd_line = key_list_to_cmd_line
        self.disconnect_from_editor()

        # For receiving key event from codeeditor
        if connect_editor:
            editor = self.get_editor()
            self.dot_cmd.editor_connected = editor
            editor.sig_key_pressed.connect(self.rcv_key_from_editor)

        # Get selection region.
        if self.is_normal():
            return

        sel_start = self.get_pos_start_in_selection()
        sel_end = self.get_pos_end_in_selection()

        block_start, block_no_start = self.cursor.get_block(sel_start)
        block_end, block_no_end = self.cursor.get_block(sel_end)

        dot_cmd.sel_block_no_start = block_no_start
        dot_cmd.sel_block_no_end = block_no_end
        dot_cmd.sel_col_start = sel_start - block_start.position()
        dot_cmd.sel_col_end = sel_end - block_end.position()

    def get_register_name(self):
        """Get register name from previous command."""
        txt = self.input_cmd_prev.cmd

        if len(txt) == 2 and txt[0] == '"':
            return txt[1]
        else:
            return '"'  # default register name

    def set_register(self, name, content, register_type):
        """Set content into register_dict."""
        if name == "+":
            QApplication.clipboard().setText(content)
        else:
            info = self.register_dict[name]
            info.name = name
            info.content = content
            info.type = register_type
            self.register_dict[name] = info

    def get_register(self):
        """Get content from register_dict."""
        name = self.get_register_name()
        if name == "+":
            info = RegisterInfo()
            info.name = "+"
            info.content = QApplication.clipboard().text()
            info.type = VimState.NORMAL
            return info
        else:
            return self.register_dict[name]

    # ---- Application helpers ----------------------------------------
    def _get_application_plugin(self):
        """Return and cache Spyder's application plugin."""
        plugin = getattr(self, "_application_plugin", None)
        if plugin is None:
            try:
                plugin = self.main.get_plugin(Plugins.Application)
            except Exception:
                return None
            self._application_plugin = plugin
        return plugin

    def _open_file_in_application(self, file_path: str) -> None:
        """Open *file_path* through the application plugin when available."""
        if not file_path:
            return
        application = self._get_application_plugin()
        open_file = getattr(application, "open_file_in_plugin", None)
        if callable(open_file):
            open_file(file_path)

    # ---- Bookmarks ----------------------------------------------------
    def _load_persistent_bookmarks(self) -> None:
        """Load persistent bookmarks from disk."""
        self.bookmark_manager._load_persistent_bookmarks()

    def _save_persistent_bookmarks(self) -> None:
        """Save persistent bookmarks to disk."""
        self.bookmark_manager._save_persistent_bookmarks()

    def set_bookmark(self, name: str) -> None:
        """Set bookmark at current cursor position."""
        self.bookmark_manager.set_bookmark(name)

    def get_bookmark(self, name: str):
        """Return bookmark information for *name* or None."""
        return self.bookmark_manager.get_bookmark(name)

    def jump_to_bookmark(self, name: str) -> None:
        """Move cursor to bookmark *name* if it exists."""
        self.bookmark_manager.jump_to_bookmark(name)

    # ---- Jump list ---------------------------------------------------
    def get_current_location(self):
        """Return the current file path and cursor position or ``None``."""
        stack = self.get_editorstack()
        editor = self.get_editor()
        if not stack or not editor or editor.textCursor().isNull():
            return None
        try:
            return stack.get_current_filename(), editor.textCursor().position()
        except Exception:
            return None

    # ---- Jump list ---------------------------------------------------
    def push_jump(self) -> None:
        """Record the current cursor position in the jump list."""
        location = self.get_current_location()
        if location is None:
            return
        file_path, pos = location
        self.jump_list.push(file_path, pos)

    def start_definition_tracking(self, previous):
        """Monitor cursor movement after ``gd`` and update the jump list."""
        timer = QTimer(self)
        self.timer_go_to_definition = timer

        def stop():
            timer.stop()
            self.timer_go_to_definition = None

        def check():
            current = self.get_current_location()
            if previous is not None and current and current != previous:
                pos = current[1]
                
                if pos < 1:
                    # The cursor may be set to 0 before it is created and positioned,
                    # but a definition never starts at cursor position 0.
                    return
                self.push_jump()
                stop()

        def finalize():
            stop()
            current = self.get_current_location()
            if previous is not None and current == previous:
                self.jump_list.pop_last()

        timer.timeout.connect(check)
        timer.start(100)
        QTimer.singleShot(2000, finalize)

    def jump_backward(self) -> None:
        """Jump to the previous location in the jump list."""
        jump = self.jump_list.back()
        if not jump:
            return
        if self.get_editorstack().is_file_opened(jump.file) is None:
            self._open_file_in_application(jump.file)
        self.get_editorstack().set_current_filename(jump.file)
        self.cursor.set_cursor_pos(jump.pos)

    def jump_forward(self) -> None:
        """Jump to the next location in the jump list."""
        jump = self.jump_list.forward()
        if not jump:
            return
        if self.get_editorstack().is_file_opened(jump.file) is None:
            self._open_file_in_application(jump.file)
        self.get_editorstack().set_current_filename(jump.file)
        self.cursor.set_cursor_pos(jump.pos)

    def set_message(self, msg, duration_ms=-1):
        """Display ``msg`` in the status bar."""
        self.msg_label.setText(f"{self.msg_prefix}{msg}")

    def start_recording_macro(self, reg_name):
        """Start recording macro."""
        self.msg_prefix = f"recording @{reg_name}... "

        editor = self.get_editor()
        self.manager_macro.start_record(reg_name)
        self.manager_macro.connect_to_editor(
            editor, self.add_key_from_editor_to_macro_manager
        )

    def is_recording_macro(self):
        """Return is_recording."""
        return self.manager_macro.is_recording

    def stop_recording_macro(self):
        """Stop recording macro."""
        self.msg_prefix = ""
        self.manager_macro.stop_record()
        self.manager_macro.disconnect_from_editor(
            self.add_key_from_editor_to_macro_manager
        )

    @Slot(QKeyEvent)
    def add_key_from_editor_to_macro_manager(self, event):
        """Add key event from editor to list to macro_manager."""
        self.manager_macro.add_editor_keyevent(event)

    def set_marker_for_easymotion(self, positions: list[int], motion_type):
        """Set markers for EasyMotion.

        Args:
            positions: Character positions to annotate.
            motion_type: Kind of motion that triggered the annotation.
        """
        if not positions:
            return
        self.remove_marker_of_easymotion()
        editor = self.get_editor()
        self.manager_marker_easymotion.set_positions(positions, motion_type)
        self.painter_easymotion.editor = editor
        self.painter_easymotion.positions = positions
        self.painter_easymotion.names = self.manager_marker_easymotion.name_list
        self.editor_connected_easymotion = editor
        editor.viewport().installEventFilter(self.painter_easymotion)
        editor.viewport().update()

    def update_marker_for_easymotion(self):
        """Update marker of easymotion."""
        self.remove_marker_of_easymotion()
        editor = self.get_editor()
        self.painter_easymotion.editor = editor
        self.painter_easymotion.positions = self.manager_marker_easymotion.position_list
        self.painter_easymotion.names = self.manager_marker_easymotion.name_list
        self.editor_connected_easymotion = editor
        editor.viewport().installEventFilter(self.painter_easymotion)
        editor.viewport().update()

    def remove_marker_of_easymotion(self):
        """Remove marker of easymotion."""
        editor = self.editor_connected_easymotion
        self.editor_connected_easymotion = None
        if editor:
            editor.viewport().removeEventFilter(self.painter_easymotion)
            editor.viewport().update()

    @Slot()
    def hide_annotate_on_txt(self):
        """Hide Labels for annotate on txt."""
        for idx in range(self.n_annotate_max):
            self.labels_for_annotate[idx].hide()

    def annotate_on_txt(self, info: dict[int, str], timeout: int = -1):
        """Annotate the editor with temporary inline labels.

        Args:
            info: Mapping of cursor positions to text labels.
            timeout: Milliseconds before labels are cleared.
        """
        editor = self.get_editor()
        tc = editor.textCursor()
        font = editor.viewport().font()
        fm = editor.fontMetrics()
        ch_width = fm.width("H")
        ch_height = fm.height()

        for idx, (pos, data) in enumerate(info.items()):
            label = self.labels_for_annotate[idx]
            label.set_style(font.family(), font.pointSize())
            label.setFixedSize(ch_width * len(data), ch_height)
            label.setText(data)

            tc.setPosition(pos)
            rect_in_local = editor.cursorRect(tc)
            pos_in_parent = editor.viewport().mapTo(
                label.parent(), rect_in_local.topLeft()
            )
            print(pos_in_parent)
            label.move(pos_in_parent)
            label.raise_()
            label.show()

            if idx > self.n_annotate_max - 1:
                break

        QTimer.singleShot(timeout, self.hide_annotate_on_txt)



# -*- coding: utf-8 -*-
"""Tests for the file search dialog."""

# Third Party Libraries
from qtpy.QtCore import QEvent, Qt
from qtpy.QtGui import QKeyEvent

# Project Libraries
from spyder_okvim.utils.file_search import FileSearchDialog
from spyder_okvim.utils.jump_list import Jump


def test_open_file_search(vim_bot, monkeypatch, tmpdir):
    """Test opening the file search dialog and jump list update."""
    main, stack, _, vim, qtbot = vim_bot

    fn = tmpdir.join("tmp.txt")
    fn.write("content")

    monkeypatch.setattr(FileSearchDialog, "exec_", lambda x: x)
    monkeypatch.setattr(FileSearchDialog, "get_selected_path", lambda x: str(fn))

    monkeypatch.setattr(main, "load_edit", lambda path: None)

    vs = vim.vim_cmd.vim_status
    vs.cursor.set_cursor_pos(0)
    vs.reset_for_test()
    current = stack.get_current_filename()

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_P, Qt.ControlModifier)
    vim.vim_cmd.commandline.keyPressEvent(event)

    assert vs.jump_list.jumps == [
        Jump(current, 0),
        Jump(str(fn), 0),
    ]


def test_invalid_folder(qtbot):
    """Test dialog behavior with an invalid folder."""
    pf = FileSearchDialog(None)
    qtbot.addWidget(pf)
    pf.show()

    edit = pf.edit

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_P, Qt.ControlModifier)
    edit.keyPressEvent(event)

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_N, Qt.ControlModifier)
    edit.keyPressEvent(event)

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_U, Qt.ControlModifier)
    edit.keyPressEvent(event)

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_D, Qt.ControlModifier)
    edit.keyPressEvent(event)

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_F, Qt.ControlModifier)
    edit.keyPressEvent(event)

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_U, Qt.ControlModifier)
    edit.keyPressEvent(event)

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_Up, Qt.NoModifier)
    edit.keyPressEvent(event)

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_Down, Qt.NoModifier)
    edit.keyPressEvent(event)

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_PageUp, Qt.NoModifier)
    edit.keyPressEvent(event)

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_PageDown, Qt.NoModifier)
    edit.keyPressEvent(event)

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_Enter, Qt.NoModifier)
    edit.keyPressEvent(event)
    assert pf.get_selected_path() == ""

    pf.show()
    event = QKeyEvent(QEvent.KeyPress, Qt.Key_Escape, Qt.NoModifier)
    edit.keyPressEvent(event)
    assert pf.get_selected_path() == ""


def test_valid_folder(qtbot, tmpdir):
    """Test valid folder."""
    folder = tmpdir.mkdir("projects")
    for name in ["a", "ok97465.py", "test.md", "fig.txt"]:
        tmp = folder.join(name)
        tmp.write("contents")

    folder_sub = folder.mkdir("sub1")
    for name in ["b", "path_finder.py", "__init__.md", "config.txt"]:
        tmp = folder_sub.join(name)
        tmp.write("contents")

    pf = FileSearchDialog(str(folder))

    qtbot.addWidget(pf)
    pf.show()

    edit = pf.edit
    lv = pf.list_viewer
    lm = pf.list_model

    assert lm.rowCount() == 6

    qtbot.keyPress(edit, Qt.Key_N, modifier=Qt.ControlModifier)
    assert lv.currentIndex().row() == 1

    qtbot.keyPress(edit, Qt.Key_P, modifier=Qt.ControlModifier)
    assert lv.currentIndex().row() == 0

    qtbot.keyPress(edit, Qt.Key_Down)
    assert lv.currentIndex().row() == 1

    qtbot.keyPress(edit, Qt.Key_Up)
    assert lv.currentIndex().row() == 0

    qtbot.keyPress(edit, Qt.Key_D, modifier=Qt.ControlModifier)
    assert lv.currentIndex().row() > 0

    qtbot.keyPress(edit, Qt.Key_U, modifier=Qt.ControlModifier)
    qtbot.keyPress(edit, Qt.Key_U, modifier=Qt.ControlModifier)
    assert lv.currentIndex().row() == 0

    qtbot.keyPress(edit, Qt.Key_F, modifier=Qt.ControlModifier)
    assert lv.currentIndex().row() > 0

    qtbot.keyPress(edit, Qt.Key_B, modifier=Qt.ControlModifier)
    qtbot.keyPress(edit, Qt.Key_B, modifier=Qt.ControlModifier)
    assert lv.currentIndex().row() == 0

    qtbot.keyPress(edit, Qt.Key_PageDown)
    assert lv.currentIndex().row() > 0

    qtbot.keyPress(edit, Qt.Key_PageUp)
    assert lv.currentIndex().row() == 0

    qtbot.keyClicks(edit, "o")
    assert lm.rowCount() == 2

    qtbot.keyPress(edit, Qt.Key_Backspace)
    assert lm.rowCount() == 6

    qtbot.keyClicks(edit, "ok9")
    qtbot.keyPress(edit, Qt.Key_Enter)
    assert pf.get_selected_path() != ""



"""Manage persistent and local bookmarks for OkVim."""

from __future__ import annotations

# Standard Libraries
import json
import os
import os.path as osp
from collections import defaultdict
from collections.abc import Callable


class BookmarkManager:
    """Handle loading, saving and retrieving bookmarks."""

    def __init__(
        self,
        bookmarks_file: str,
        open_file_func: Callable[[str], None],
        get_editorstack_func: Callable[[], object],
        get_editor_func: Callable[[], object],
        set_cursor_pos_func: Callable[[int], None],
    ) -> None:
        self.bookmarks_file = bookmarks_file
        self.open_file = open_file_func
        self.get_editorstack = get_editorstack_func
        self.get_editor = get_editor_func
        self.set_cursor_pos = set_cursor_pos_func

        self.bookmarks: dict[str, dict[str, dict[str, int]]] = defaultdict(dict)
        self.bookmarks_global: dict[str, dict[str, int]] = {}
        self._load_persistent_bookmarks()

    # ------------------------------------------------------------------
    # Persistence helpers
    # ------------------------------------------------------------------
    def _load_persistent_bookmarks(self) -> None:
        """Load persistent bookmarks from disk if present."""
        if osp.isfile(self.bookmarks_file):
            try:
                with open(self.bookmarks_file, "r", encoding="utf-8") as fh:
                    self.bookmarks_global = json.load(fh)
            except Exception:
                self.bookmarks_global = {}
        else:
            self.bookmarks_global = {}

    def _save_persistent_bookmarks(self) -> None:
        """Save persistent bookmarks to disk."""
        folder = osp.dirname(self.bookmarks_file)
        os.makedirs(folder, exist_ok=True)
        with open(self.bookmarks_file, "w", encoding="utf-8") as fh:
            json.dump(self.bookmarks_global, fh)

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------
    def clear(self) -> None:
        """Remove all bookmarks from memory and disk."""
        self.bookmarks = defaultdict(dict)
        self.bookmarks_global = {}
        self._save_persistent_bookmarks()

    def set_bookmark(self, name: str) -> None:
        """Set bookmark at the current cursor position."""
        editor = self.get_editor()
        cursor = editor.textCursor()
        line = cursor.blockNumber()
        col = cursor.position() - cursor.block().position()
        path = self.get_editorstack().get_current_filename()
        info = {"file": path, "line": line, "col": col}
        if name.isupper():
            self.bookmarks_global[name] = info
            self._save_persistent_bookmarks()
        else:
            self.bookmarks[path][name] = info

    def get_bookmark(self, name: str):
        """Return bookmark information for *name* or None."""
        if name.isupper():
            return self.bookmarks_global.get(name)
        current = self.get_editorstack().get_current_filename()
        return self.bookmarks.get(current, {}).get(name)

    def jump_to_bookmark(self, name: str) -> None:
        """Move cursor to bookmark *name* if it exists."""
        if name.isupper():
            info = self.bookmarks_global.get(name)
            if not info:
                return
            file_path = info.get("file")
            line = info.get("line")
            col = info.get("col")
            editor_stack = self.get_editorstack()
            if editor_stack.is_file_opened(file_path) is None:
                self.open_file(file_path)
            editor_stack.set_current_filename(file_path)
            editor = self.get_editor()
            block = editor.document().findBlockByNumber(line)
            if not block.isValid():
                self.bookmarks_global.pop(name, None)
                self._save_persistent_bookmarks()
                return
            pos = block.position() + min(col, block.length() - 1)
            self.set_cursor_pos(pos)
        else:
            current = self.get_editorstack().get_current_filename()
            info = self.bookmarks.get(current, {}).get(name)
            if not info:
                return
            line = info.get("line")
            col = info.get("col")
            editor = self.get_editor()
            block = editor.document().findBlockByNumber(line)
            if not block.isValid():
                self.bookmarks[current].pop(name, None)
                return
            pos = block.position() + min(col, block.length() - 1)
            self.set_cursor_pos(pos)

    def list_bookmarks(self) -> list[tuple[str, dict]]:
        """Return a combined list of local and global bookmarks."""
        current = self.get_editorstack().get_current_filename()
        marks = []
        for name, info in sorted(self.bookmarks.get(current, {}).items()):
            marks.append((name, info))
        for name, info in sorted(self.bookmarks_global.items()):
            marks.append((name, info))
        return marks



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2021, Spyder Bot
#
# Licensed under the terms of the MIT license
# ----------------------------------------------------------------------------
"""Spyder Custom Layout API."""

# Third Party Libraries
from spyder.api.plugins import Plugins
from spyder.plugins.layout.api import BaseGridLayoutType


class SpyderCustomLayouts:
    CustomLayout = "spyder_custom_layout"


class CustomLayout(BaseGridLayoutType):
    """Custom layout for the SpyderCustomLayout plugin."""

    ID = SpyderCustomLayouts.CustomLayout

    def __init__(self, parent_plugin: Plugins) -> None:
        super().__init__(parent_plugin)

        # Explorer/Files and OutlineExplorer tabbed
        # self.add_area(
        #     [Plugins.OutlineExplorer, Plugins.Explorer],
        #     0, 0, row_span=2, visible=True, default=True)

        # Explorer/Files and OutlineExplorer divided
        self.add_area([Plugins.Explorer], 0, 0, visible=True, default=True)
        self.add_area([Plugins.OutlineExplorer], 1, 0, visible=True)

        # Triggers and error since there is an unassigned space between the
        # Editor (column 1) and the VariableExplorer (column 3)
        # self.add_area([Plugins.Editor], 0, 1)

        # Editor in the middle taken at least half of the space available
        self.add_area([Plugins.Editor], 0, 1, col_span=2, row_span=2)

        # VariableExplorer in the usual place (top-right)
        self.add_area([Plugins.VariableExplorer], 0, 3)

        # Triggers and error when loading the layout since the area will
        # overlap with the Editor (col_span=2 therefore column 1 and 2 will be
        # used by the Editor)
        # self.add_area([Plugins.IPythonConsole], 1, 2)

        # Valid area definition to put the IPython Console in the default place
        # bottom-right
        self.add_area([Plugins.IPythonConsole], 1, 3)

    @staticmethod
    def get_name() -> str:
        return "Spyder_okvim"



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""OkVim Plugin."""

# Third Party Libraries
import qtawesome as qta
from qtpy.QtCore import Qt, Signal, QCoreApplication
from qtpy.QtGui import QKeySequence
from qtpy.QtWidgets import QHBoxLayout, QShortcut
from spyder.api.plugin_registration.decorators import on_plugin_available
from spyder.api.plugins import Plugins, SpyderDockablePlugin

# Project Libraries
from spyder_okvim.utils.testing_env import running_in_pytest
from spyder.api.widgets.status import StatusBarWidget
from spyder.utils.icon_manager import MAIN_FG_COLOR

# Project Libraries
from spyder_okvim.spyder.api import CustomLayout
from spyder_okvim.spyder.config import CONF_DEFAULTS, CONF_SECTION, CONF_VERSION
from spyder_okvim.spyder.confpage import OkvimConfigPage
from spyder_okvim.spyder.vim_widgets import VimPane


class StatusBarVimWidget(StatusBarWidget):
    """Status bar widget for okvim."""

    ID = f"{CONF_SECTION}_status_bar"

    def __init__(self, parent, msg_label, status_label, cmd_line):
        """Initialize status bar widget.

        Args:
            parent: Parent widget for this status bar widget.
            msg_label: Label used to display messages.
            status_label: Label showing the current Vim mode.
            cmd_line: Line edit used for entering commands.
        """
        self.msg_label = msg_label
        self.status_label = status_label
        self.cmd_line = cmd_line
        super().__init__(parent, show_icon=False, show_label=False)

    # ---- Private API -------------------------------------------------
    def _set_layout(self):
        """Create the internal layout of the status bar widget."""
        spacing_post = 12
        spacing = 5

        self.msg_label.setMinimumWidth(90)
        self.cmd_line.setMinimumWidth(80)
        width_msg = self.msg_label.sizeHint().width()
        width_status = self.status_label.sizeHint().width()
        width_cmd = self.cmd_line.sizeHint().width()

        width_total = width_msg + width_status + width_cmd

        layout = QHBoxLayout(self)
        layout.setSpacing(spacing)
        layout.addWidget(self.msg_label, int(width_msg / width_total * 100))
        layout.addWidget(self.status_label, int(width_status / width_total * 100))
        layout.addWidget(self.cmd_line, int(width_cmd / width_total * 100))
        layout.addSpacing(spacing_post)
        layout.setContentsMargins(0, 0, 0, 0)

        width_total += 2 * spacing + spacing_post
        self.setFixedWidth(width_total)
        self.setLayout(layout)

    def set_layout(self):
        """Public entry point to apply the layout."""
        self._set_layout()

    # ---- Status bar widget API
    def set_icon(self):
        """Set the icon for the status bar widget."""
        pass

    def set_value(self, value):
        """Set formatted text value."""
        pass

    def update_tooltip(self):
        """Update tooltip for widget."""
        pass

    def mouseReleaseEvent(self, event):
        """Override Qt method to allow for click signal."""
        super().mouseReleaseEvent(event)

    # ---- API to be defined by user
    def get_tooltip(self):
        """Return the widget tooltip text."""
        return ""

    def get_icon(self):
        """Return the widget icon."""
        return None


class OkVim(SpyderDockablePlugin):  # pylint: disable=R0904
    """Implements a Vim-like command mode."""

    focus_changed = Signal()
    NAME = CONF_SECTION
    REQUIRES = [
        Plugins.StatusBar,
        Plugins.Preferences,
        Plugins.Application,
        Plugins.Editor,
    ]
    OPTIONAL = []
    WIDGET_CLASS = VimPane
    CONF_SECTION = CONF_SECTION
    CONF_WIDGET_CLASS = OkvimConfigPage
    CONF_DEFAULTS = CONF_DEFAULTS
    CONF_ = CONF_VERSION
    CUSTOM_LAYOUTS = [CustomLayout]
    CAN_BE_DISABLED = True
    RAISE_AND_FOCUS = True

    @property
    def vim_cmd(self):
        """Return vim_cmd for test."""
        return self.get_widget().vim_cmd

    # --- SpyderDockablePlugin API
    # ------------------------------------------------------------------------
    @staticmethod
    def get_name() -> str:
        """Return name."""
        return CONF_SECTION

    @staticmethod
    def get_description() -> str:
        """Return description."""
        return "A plugin for Spyder to enable Vim keybindings"

    @classmethod
    def get_icon(cls):
        """Return widget icon."""
        return qta.icon("mdi.vimeo", color=MAIN_FG_COLOR)

    def on_initialize(self) -> None:
        """Perform plugin initialization after it is added to Spyder."""
        vim_cmd = self.get_widget().vim_cmd

        status_bar_widget = StatusBarVimWidget(
            self._main,
            vim_cmd.msg_label,
            vim_cmd.status_label,
            vim_cmd.commandline,
        )

        statusbar = self.get_plugin(Plugins.StatusBar)
        statusbar.add_status_widget(status_bar_widget)

        editorsplitter = vim_cmd.editor_widget.get_widget().editorsplitter

        esc_shortcut = QShortcut(
            QKeySequence("Esc"),
            editorsplitter,
            vim_cmd.commandline.setFocus,
        )
        esc_shortcut.setContext(Qt.WidgetWithChildrenShortcut)

    @on_plugin_available(plugin=Plugins.Preferences)
    def on_preferences_available(self) -> None:
        """Connect when preferences available."""
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.register_plugin_preferences(self)

    @staticmethod
    def check_compatibility():
        """Check plugin compatibility."""
        valid = True
        message = ""  # Note: Remember to use _("") to localize the string
        return valid, message

    def on_close(self, cancellable: bool = True) -> bool:
        """Handle plugin shutdown."""
        widget = self.get_widget()
        if widget is not None:
            try:
                widget.vim_cmd.cleanup()
            except AttributeError:
                pass
            widget.close()
            if running_in_pytest():
                widget.deleteLater()
        QCoreApplication.processEvents()
        return True

    def get_plugin_actions(self):
        """Return plugin actions."""
        return []

    # ------ SpyderPluginMixin API
    def apply_plugin_settings(self, options) -> None:
        """Apply the config settings."""
        self.get_widget().apply_plugin_settings(options)



# -*- coding: utf-8 -*-
"""Dialog for locating files inside a project tree."""
from __future__ import annotations

# Standard Libraries
import os.path as osp
import re
import sys

# Third Party Libraries
from qtpy.QtCore import QDir, QDirIterator, Qt, Signal
from qtpy.QtGui import QKeyEvent, QStandardItem
from qtpy.QtWidgets import QApplication, QDialog, QLineEdit, QWidget
from spyder.config.gui import get_font

from .list_dialog import PopupTableDialog


def fuzzyfinder(query: str, collection: list[str]) -> list[str]:
    """Fuzzy search.

    Ref: https://github.com/amjith/fuzzyfinder

    Args:
        query: A partial string typically entered by a user.
        collection: Collection of strings filtered by ``query``.

    Returns:
        list: Suggestions narrowed down from ``collection`` using ``query``.

    """
    suggestions = []
    pat = ".*?".join(map(re.escape, query))
    # lookahead regex to manage overlapping matches
    pat = "(?=({0}))".format(pat)
    regex = re.compile(pat, re.IGNORECASE)

    for txt in collection:
        r = list(regex.finditer(txt))
        if r:
            best = min(r, key=lambda x: len(x.group(1)))  # find shortest match
            suggestions.append((len(best.group(1)), txt))

    return [z[-1] for z in sorted(suggestions)]


class FileSearchLineEdit(QLineEdit):
    """Line editor used by :class:`FileSearchDialog`."""

    sig_up_key_pressed = Signal()
    sig_pg_up_key_pressed = Signal()
    sig_pg_half_up_key_pressed = Signal()
    sig_down_key_pressed = Signal()
    sig_pg_down_key_pressed = Signal()
    sig_pg_half_down_key_pressed = Signal()
    sig_enter_key_pressed = Signal()
    sig_esc_key_pressed = Signal()

    def __init__(self, parent: QDialog | None, textChanged) -> None:
        """Initialize the line edit.

        Args:
            parent: Owning dialog.
            textChanged: Slot to invoke when text changes.
        """
        super().__init__(parent, textChanged=textChanged)
        self.dispatcher_nomodifier = {
            Qt.Key_Up: self.sig_up_key_pressed.emit,
            Qt.Key_Down: self.sig_down_key_pressed.emit,
            Qt.Key_Enter: self.sig_enter_key_pressed.emit,
            Qt.Key_Return: self.sig_enter_key_pressed.emit,
            Qt.Key_Escape: self.sig_esc_key_pressed.emit,
            Qt.Key_PageUp: self.sig_pg_up_key_pressed.emit,
            Qt.Key_PageDown: self.sig_pg_down_key_pressed.emit,
        }

        self.dispatcher_ctrl = {
            Qt.Key_P: self.sig_up_key_pressed.emit,
            Qt.Key_B: self.sig_pg_up_key_pressed.emit,
            Qt.Key_U: self.sig_pg_half_up_key_pressed.emit,
            Qt.Key_N: self.sig_down_key_pressed.emit,
            Qt.Key_F: self.sig_pg_down_key_pressed.emit,
            Qt.Key_D: self.sig_pg_half_down_key_pressed.emit,
        }

    def keyPressEvent(self, e: QKeyEvent) -> None:
        """Override Qt method."""
        key = e.key()
        modifier = e.modifiers()
        pressed_ctrl = modifier == Qt.ControlModifier
        pressed_nomodifier = modifier == Qt.NoModifier

        if pressed_nomodifier and key in self.dispatcher_nomodifier.keys():
            self.dispatcher_nomodifier[key]()
            return

        if pressed_ctrl and key in self.dispatcher_ctrl.keys():
            self.dispatcher_ctrl[key]()
            return

        super().keyPressEvent(e)


class FileSearchDialog(PopupTableDialog):
    """Dialog used to select a file path from a project."""

    _MIN_WIDTH = 800
    _MAX_HEIGHT = 600

    def __init__(self, folder: str | None, parent: QWidget | None = None) -> None:
        """Create the dialog and populate file information.

        Args:
            folder: Root directory to search.
            parent: Parent widget for the dialog.
        """
        super().__init__(
            "Path Finder",
            parent=parent,
            headers=["File", "Folder"],
            min_width=self._MIN_WIDTH,
            max_height=self._MAX_HEIGHT,
        )
        font = get_font(font_size_delta=2)

        self.folder = folder
        self.path_selected = ""
        self.path_list = []
        self.results_old = {}
        self.list_viewer.setAttribute(Qt.WA_TransparentForMouseEvents)
        self.list_viewer.setFocusPolicy(Qt.NoFocus)

        # Set edit
        self.edit = FileSearchLineEdit(self, textChanged=self.update_list)
        self.edit.setFont(font)
        self.edit.sig_esc_key_pressed.connect(self.close)
        self.edit.sig_enter_key_pressed.connect(self.enter)

        self.edit.sig_up_key_pressed.connect(lambda: self.prev_row(1))
        self.edit.sig_pg_up_key_pressed.connect(self.pg_up)
        self.edit.sig_pg_half_up_key_pressed.connect(self.pg_half_up)

        self.edit.sig_down_key_pressed.connect(lambda: self.next_row(1))
        self.edit.sig_pg_down_key_pressed.connect(self.pg_down)
        self.edit.sig_pg_half_down_key_pressed.connect(self.pg_half_down)

        self.layout_.insertWidget(0, self.edit)

        self.collect_paths()
        self.update_list()

        self.edit.setFocus()

    def get_selected_path(self) -> str:
        """Return the file path chosen by the user."""
        return self.path_selected

    def collect_paths(self) -> None:
        """Populate the list of discoverable file paths."""
        if self.folder is None or not osp.isdir(self.folder):
            self.edit.setPlaceholderText("The project is not valid.")
        else:
            dir_ = QDir(self.folder)
            it = QDirIterator(
                self.folder,
                ["*.py", "*.txt", "*.md"],
                QDir.Files | QDir.NoDotAndDotDot | QDir.NoSymLinks,
                QDirIterator.Subdirectories,
            )
            while it.hasNext():
                self.path_list.append(dir_.relativeFilePath(it.next()))

        self.results_old[""] = self.path_list

    def update_list(self) -> None:
        """Update listview."""
        query = self.edit.text()
        query = query.replace(" ", "")
        paths = self.results_old.get(query, None)

        if paths is None:
            collections = self.results_old.get(query[:-1], self.path_list)
            paths = fuzzyfinder(query, collections)

        if paths:
            self.list_model.setRowCount(0)
            for path in paths:
                basename = osp.basename(path)
                dirname = osp.dirname(path)
                item = QStandardItem(basename)
                item.setData(path, Qt.UserRole)
                row = [item, QStandardItem(dirname)]
                for it in row:
                    it.setEditable(False)
                self.list_model.appendRow(row)
            self.list_viewer.setCurrentIndex(self.list_model.index(0, 0))
            self.list_viewer.selectRow(0)

    def enter(self) -> None:
        """Select next row in list viewer."""
        idx = self.list_viewer.currentIndex()
        item = self.list_model.item(idx.row(), 0)
        rel_path = item.data(Qt.UserRole) if item else None
        if rel_path and isinstance(self.folder, str):
            path = osp.join(self.folder, rel_path)
            self.path_selected = path
        self.close()



# -*- coding: utf-8 -*-
"""Support for EasyMotion-like navigation.

EasyMotion highlights possible jump targets in the editor and lets the user
select one with a short key sequence.  This executor provides helper methods
used by the normal and visual mode executors to implement those jumps.
"""

# Third Party Libraries
from qtpy.QtCore import QPoint
from qtpy.QtGui import QTextCursor, QTextDocument

# Project Libraries
from spyder_okvim.executor.executor_base import (
    RETURN_EXECUTOR_METHOD_INFO,
    ExecutorSubBase,
)
from spyder_okvim.utils.motion import MotionInfo, MotionType


class ExecutorSelectMarkerEasymotion(ExecutorSubBase):
    """Submode for selecting marker of easymotion."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False
        self.motion_info = MotionInfo()

    def _set_motion_info(
        self,
        cur_pos: int | None,
        sel_start: int | None = None,
        sel_end: int | None = None,
        motion_type: int = MotionType.LineWise,
    ):
        """Set motion info.

        Args:
            cur_pos: the position of cursor.
            sel_start: the start position of selection.
            sel_end: the end position of selection.
            motion_type: motion type

        Returns:
            MotionInfo

        """
        self.motion_info.cursor_pos = cur_pos
        self.motion_info.sel_start = sel_start
        self.motion_info.sel_end = sel_end
        self.motion_info.motion_type = motion_type

        return self.motion_info

    def __call__(self, txt: str):
        manager_marker = self.vim_status.manager_marker_easymotion
        manager_marker.handle_user_input(txt)

        motion_type = manager_marker.motion_type
        pos_list = manager_marker.position_list
        len_pos = len(pos_list)

        if len_pos > 1:
            self.vim_status.update_marker_for_easymotion()
            return True
        elif len_pos == 0:
            self.vim_status.remove_marker_of_easymotion()
            self.vim_status.sub_mode = None
            return True
        else:
            self.vim_status.remove_marker_of_easymotion()
            self.vim_status.sub_mode = None
            self._set_motion_info(pos_list[0], motion_type=motion_type)
            return self.process_return(self.execute_func_deferred(self.motion_info))


class ExecutorSearchCharEasymotion(ExecutorSubBase):
    """Submode for searching character in easymotion."""

    def __init__(self, vim_status, executor_select_maker):
        super().__init__(vim_status)
        self.allow_leaderkey = False
        self.executor_select_maker = executor_select_maker
        self.n_input_required = 1
        self.search_method = None

    def __call__(self, txt: str):
        if len(txt) < self.n_input_required:
            return False

        self.vim_status.sub_mode = None
        ret = None
        if self.search_method:
            ret = self.search_method(txt)

        if ret:
            self.vim_status.sub_mode = ret.sub_mode
            return ret.clear_command_line
        else:
            self.vim_status.sub_mode = None
            return True

    def get_cursor_pos_of_viewport(self) -> tuple[int, int]:
        """Get the cursor position of viewport of editor."""
        editor = self.vim_status.get_editor()
        start_pos = editor.cursorForPosition(QPoint(0, 0)).position()
        bottom_right = QPoint(
            editor.viewport().width() - 1, editor.viewport().height() - 1
        )
        end_pos = editor.cursorForPosition(bottom_right).position()

        return start_pos, end_pos

    def set_search_method(self, method_name: str):
        "Set search method."
        self.search_method = {
            "ch_forwards": self.forwards_find_char,
            "ch_backwards": self.backwards_find_char,
        }.get(method_name, None)

    def set_position_result_to_vim_status(self, positions, motion_type):
        """Set positions to vim status for easymotion."""
        executor_sub = self.executor_select_maker
        if positions:
            executor_sub.set_func_list_deferred(
                self.func_list_deferred, self.return_deferred
            )
            self.vim_status.set_marker_for_easymotion(positions, motion_type)
            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def forwards_find_char(self, characters: str):
        """Find characters forwards."""
        editor = self.vim_status.get_editor()

        cur_pos = editor.textCursor().position()
        view_start_pos, view_end_pos = self.get_cursor_pos_of_viewport()
        start_pos = min([view_end_pos, cur_pos]) + 1

        positions = []
        while view_start_pos <= start_pos <= view_end_pos:
            cursor = editor.document().find(characters, start_pos)
            if cursor.isNull():
                break
            positions.append(cursor.position() - len(characters))
            start_pos = cursor.position()

        return self.set_position_result_to_vim_status(
            positions, MotionType.CharWiseIncludingEnd
        )

    def backwards_find_char(self, characters: str):
        """Find characters backwards."""
        editor = self.vim_status.get_editor()

        cur_pos = editor.textCursor().position()
        view_start_pos, view_end_pos = self.get_cursor_pos_of_viewport()
        start_pos = max([view_start_pos, cur_pos]) - 1

        positions = []
        while view_start_pos <= start_pos <= view_end_pos:
            cursor = editor.document().find(
                characters, start_pos, QTextDocument.FindBackward
            )
            if cursor.isNull():
                break
            positions.append(cursor.position() - len(characters))
            start_pos = cursor.position() - len(characters)

        return self.set_position_result_to_vim_status(positions, MotionType.CharWise)


class ExecutorEasymotion(ExecutorSubBase):
    """Submode of easymotion."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

        self.has_zero_cmd = False

        self.dispatcher = {
            "w": self.forward_words,
            "b": self.backward_words,
            "j": self.forward_start_of_line,
            "k": self.backward_start_of_line,
            "f": self.forwards_find_char,
            "F": self.backwards_find_char,
        }
        self.executor_select_maker = ExecutorSelectMarkerEasymotion(vim_status)
        self.executor_search_char = ExecutorSearchCharEasymotion(
            vim_status, self.executor_select_maker
        )

    def __call__(self, txt: str):
        if txt.isdigit():
            return False
        num = 1
        num_str = ""
        if len(txt) > 1:
            num_str = txt[:-1]
            num = int(num_str)

        method = self.dispatcher.get(txt[-1], None)

        ret = None
        if method:
            ret = method(num, num_str)

        if ret:
            self.vim_status.sub_mode = ret.sub_mode
            return ret.clear_command_line
        else:
            self.vim_status.sub_mode = None
            return True

    def get_cursor_pos_of_viewport(self) -> tuple[int, int]:
        """Get the cursor position of viewport of editor."""
        editor = self.vim_status.get_editor()
        start_pos = editor.cursorForPosition(QPoint(0, 0)).position()
        bottom_right = QPoint(
            editor.viewport().width() - 1, editor.viewport().height() - 1
        )
        end_pos = editor.cursorForPosition(bottom_right).position()

        return start_pos, end_pos

    def set_position_result_to_vim_status(self, positions, motion_type):
        """Set positions to vim status for easymotion."""
        executor_sub = self.executor_select_maker
        if positions:
            executor_sub.set_func_list_deferred(
                self.func_list_deferred,
                self.return_deferred,
            )
            self.vim_status.set_marker_for_easymotion(positions, motion_type)
            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def forward_words(self, num=1, num_str=""):
        """Easymotion-w."""
        editor = self.vim_status.get_editor()
        cur_pos = editor.textCursor().position()
        view_start_pos, view_end_pos = self.get_cursor_pos_of_viewport()
        start_pos = max([view_start_pos, cur_pos])

        cursor = editor.textCursor()
        cursor.setPosition(start_pos)
        texts = editor.toPlainText()

        positions = []
        while 1:
            cursor.movePosition(QTextCursor.NextWord)
            if cursor.position() >= view_end_pos:
                break
            if cursor.atBlockEnd():
                continue
            ch = texts[cursor.position()]
            if not ch.isalnum():
                continue

            positions.append(cursor.position())

        return self.set_position_result_to_vim_status(positions, MotionType.CharWise)

    def backward_words(self, num=1, num_str=""):
        """Easymotion-b."""
        editor = self.vim_status.get_editor()
        cur_pos = editor.textCursor().position()
        view_start_pos, view_end_pos = self.get_cursor_pos_of_viewport()
        start_pos = min([view_end_pos, cur_pos])

        cursor = editor.textCursor()
        cursor.setPosition(start_pos)
        texts = editor.toPlainText()

        positions = []
        while 1:
            if cursor.movePosition(QTextCursor.PreviousWord) is False:
                break
            if cursor.atBlockEnd():
                continue
            ch = texts[cursor.position()]
            if not ch.isalnum():
                continue

            positions.append(cursor.position())
            if cursor.position() <= view_start_pos:
                break

        return self.set_position_result_to_vim_status(positions, MotionType.CharWise)

    def forward_start_of_line(self, num=1, num_str=""):
        """Easymotion-j."""
        editor = self.vim_status.get_editor()
        cur_pos = editor.textCursor().position()
        view_start_pos, view_end_pos = self.get_cursor_pos_of_viewport()
        start_pos = max([view_start_pos, cur_pos])

        cursor = editor.textCursor()
        cursor.setPosition(start_pos)

        positions = []
        while 1:
            if cursor.movePosition(QTextCursor.NextBlock) is False:
                break
            if cursor.position() >= view_end_pos:
                break
            pos = cursor.position()
            text = cursor.block().text()

            if text.strip():
                pos += len(text) - len(text.lstrip())

            positions.append(pos)

        return self.set_position_result_to_vim_status(positions, MotionType.LineWise)

    def backward_start_of_line(self, num=1, num_str=""):
        """Easymotion-j."""
        editor = self.vim_status.get_editor()
        cur_pos = editor.textCursor().position()
        view_start_pos, view_end_pos = self.get_cursor_pos_of_viewport()
        start_pos = min([view_end_pos, cur_pos])

        cursor = editor.textCursor()
        cursor.setPosition(start_pos)

        positions = []
        while 1:
            if cursor.movePosition(QTextCursor.PreviousBlock) is False:
                break
            if cursor.position() < view_start_pos:
                break
            pos = cursor.position()
            text = cursor.block().text()

            if text.strip():
                pos += len(text) - len(text.lstrip())

            positions.append(pos)

        return self.set_position_result_to_vim_status(positions, MotionType.LineWise)

    def forwards_find_char(self, num=1, num_str=""):
        """Find characters forwards."""
        executor_sub = self.executor_search_char
        executor_sub.set_search_method("ch_forwards")
        executor_sub.n_input_required = num
        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def backwards_find_char(self, num=1, num_str=""):
        """Find characters forwards."""
        executor_sub = self.executor_search_char
        executor_sub.set_search_method("ch_backwards")
        executor_sub.n_input_required = num
        executor_sub.set_func_list_deferred(
            self.func_list_deferred, self.return_deferred
        )
        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)
